#!/usr/bin/env python
import pandas as pd, argparse, numpy as np

def main():
	# set up command line arguments
	parser = argparse.ArgumentParser(description="Read a Table of NASA Reverb Data, return the URLs of the browse images it points to.", formatter_class=argparse.RawDescriptionHelpFormatter)

	parser.add_argument('csvfile', action='store', type=str, default='.', help='name of csv file to read in')
	parser.add_argument('--max_cloud', action='store', type=float, default=100.0, help='maximum percent cloud cover for images [100]')
	args = parser.parse_args()

	mydata = pd.read_csv(args.csvfile)

	for index, row in mydata.iterrows():
		try:
			if float(row['Cloud Cover']) <= args.max_cloud:
				try:
					print row['Browse URLs'].split(',')[0]
				except AttributeError:
					continue
			elif np.isnan(float(row['Cloud Cover'])):
				try:
					print row['Browse URLs'].split(',')[0]
				except AttributeError:
					continue
		except ValueError:
			continue

if __name__ == "__main__":
    main()



import random
import datetime as dt
import ogr
import osr
import numpy as np
from shapely.geometry import Point


def standard_landsat(instring):
    strsplit = instring.split('_')
    if len(strsplit) < 3:  # if we only have 1 (or fewer) underscores, it's already fine
        return strsplit[0]
    else:
        outstring = 'L'
        sensor = strsplit[0][1]
        # figure out what sensor we're using, and generate the appropriate string
        if sensor == '5':
            outstring += 'T5'
        elif sensor == '4':
            outstring += 'T4'
        elif sensor == '7':
            outstring += 'E7'
        elif sensor == '8':  # chances are, we won't ever see this one.
            outstring += 'C8'
        outstring += strsplit[0][2:]  # this is the path/row information
        # now it gets fun: getting the date right.
        year = strsplit[1][3:7]
        month = strsplit[1][7:9]
        day = strsplit[1][9:]
        # make sure we have 3 digits here
        doy = str(dt.datetime(int(year), int(month), int(day)).timetuple().tm_yday).zfill(3)
        outstring += year + doy
        # now, spoof the last bits so it's got the right size:
        outstring += 'XXX01'
        return outstring


def doy2mmdd(year, doy, string_out=True, outform='%Y/%m/%d'):
    datestruct = dt.datetime(year, 1, 1) + dt.timedelta(doy-1)
    if string_out:
        return datestruct.strftime(outform)
    else:
        return datestruct


def mmdd2doy(year, month, day, string_out=True):
    doy = dt.datetime(year, month, day).timetuple().tm_yday
    if string_out:
        return str(doy)
    else:
        return doy


def parse_lsat_scene(scenename, string_out=True):
    sensor = scenename[0:3]
    path = int(scenename[3:6])
    row = int(scenename[6:9])
    year = int(scenename[9:13])
    doy = int(scenename[13:16])

    if string_out:
        return sensor, str(path), str(row), str(year), str(doy)
    else:
        return sensor, path, row, year, doy


def bin_data(bins, data2bin, bindata, mode='mean'):
    digitized = np.digitize(bindata, bins)
    if mode == 'mean':
        binned = [np.nanmean(data2bin[np.logical_and(np.isfinite(bindata), digitized == i)]) for i in range(len(bins))]
    elif mode == 'median':
        binned = [np.nanmedian(data2bin[np.logical_and(np.isfinite(bindata),
                  digitized == i)]) for i in range(len(bins))]
    elif mode == 'std':
        binned = [np.nanstd(data2bin[np.logical_and(np.isfinite(bindata), digitized == i)]) for i in range(len(bins))]
    else:
        raise ValueError('mode must be mean, median, or std')
    return np.array(binned)


def random_points_in_polygon(poly, npts=1):
    xmin, ymin, xmax, ymax = poly.bounds
    rand_points = []

    for pt in range(npts):
        thisptin = False
        while not thisptin:
            rand_point = Point(xmin + (random.random() * (xmax-xmin)), ymin + (random.random() * (ymax-ymin)))
            thisptin = poly.contains(rand_point)
        rand_points.append(rand_point)

    return rand_points


def reproject_layer(inLayer, targetSRS):
    srcSRS = inLayer.GetSpatialRef()
    coordTrans = osr.CoordinateTransformation(srcSRS, targetSRS)

    outLayer = ogr.GetDriverByName('Memory').CreateDataSource('').CreateLayer(inLayer.GetName(),
                                                                              geom_type=inLayer.GetGeomType())
    inLayerDefn = inLayer.GetLayerDefn()
    for i in range(0, inLayerDefn.GetFieldCount()):
        fieldDefn = inLayerDefn.GetFieldDefn(i)
        outLayer.CreateField(fieldDefn)

    outLayerDefn = outLayer.GetLayerDefn()
    inFeature = inLayer.GetNextFeature()
    while inFeature:
        geom = inFeature.GetGeometryRef()
        geom.Transform(coordTrans)
        outFeature = ogr.Feature(outLayerDefn)
        outFeature.SetGeometry(geom)
        for i in range(0, outLayerDefn.GetFieldCount()):
            outFeature.SetField(outLayerDefn.GetFieldDefn(i).GetNameRef(), inFeature.GetField(i))

        outLayer.CreateFeature(outFeature)
        outFeature.Destroy()
        inFeature.Destroy()
        inFeature = inLayer.GetNextFeature()

    return outLayer


def round_down(num, divisor):
    return num - (num % divisor)



#!/usr/bin/env python
import argparse
import numpy as np
from pybob.coreg_tools import *


def main():
    np.seterr(all='ignore')

    # add master, slave, masks to argparse
    # can also add output directory
    parser = argparse.ArgumentParser(description="Iteratively calculate co-registration \
                                     parameters for two DEMs, as seen in Nuth and Kaeaeb (2011).")
    parser.add_argument('masterdem', type=str, help='path to master DEM to be used for co-registration')
    parser.add_argument('slavedem', type=str, help='path to slave DEM to be co-registered')
    parser.add_argument('-a', '--mask1', type=str, default=None,
                        help='Glacier mask. Areas inside of this shapefile will not be used for coregistration [None]')
    parser.add_argument('-b', '--mask2', type=str, default=None,
                        help='Land mask. Areas outside of this mask (i.e., water) \
                             will not be used for coregistration. [None]')
    parser.add_argument('-o', '--outdir', type=str, default='.',
                        help='Directory to output files to (creates if not already present). [.]')
    args = parser.parse_args()

    master, coreg_slave = dem_coregistration(args.masterdem, args.slavedem, glaciermask=args.mask1, landmask=args.mask2, outdir=args.outdir)


if __name__ == "__main__":
    main()



import os
from scipy.interpolate import griddata, interp1d
from scipy.integrate import odeint
from shapely.geometry.polygon import Polygon
from shapely.geometry.linestring import LineString
from shapely.geometry.point import Point
from pybob.GeoImg import GeoImg
from pybob.VelField import VelField
import numpy as np
import matplotlib.pyplot as plt
import geopandas as gpd
import pandas as pd


# Cell class defines a cell over which we are evaluating flux divergence,
# and thus calculating thickness. Includes normal vectors at upstream and
# downstream boundaries, as well as midpoints of upstream and downstream
# boundaries (eventually, these are x,y coordinates tied to thickness values)
class Cell(Polygon):

    def __init__(self, Flow1, Flow2):
        Coords = Flow1 + list(reversed(Flow2))
        Polygon.__init__(self, Coords)

        self.x = self.centroid.x
        self.y = self.centroid.y
        self.n_up = norm_vector(np.array(Flow1[0]) - np.array(Flow2[0]))
        self.n_dn = norm_vector(np.array(Flow2[-1]) - np.array(Flow1[-1]))

        self.x_up = (Flow2[0][0] + Flow1[0][0])/2
        self.x_dn = (Flow2[-1][0] + Flow1[-1][0])/2
        self.y_up = (Flow2[0][1] + Flow1[0][1])/2
        self.y_dn = (Flow2[-1][1] + Flow1[-1][1])/2

    def display(self, fig=None):

        if fig is None:
            fig = plt.figure()
        else:
            fig.hold(True)  # make sure we don't do away with anything

        # first, plot the boundary of the cell
        plt.plot(self.boundary.xy[0], self.boundary.xy[1], color='b', linewidth=1.5)
        plt.quiver(self.x_up, self.y_up, self.n_up[0], self.n_up[1], color='k')
        plt.quiver(self.x_dn, self.y_dn, self.n_dn[0], self.n_dn[1], color='r')
        plt.plot(self.x, self.y, 'r*', markersize=12)


# ditto for 'flowline'
class FlowLine(LineString):

    def __init__(self, Coords):
        LineString.__init__(self, Coords)
        self.x, self.y = self.xy
        self.linedist = [self.project(Point(p)) for p in self.coords]


# load spatially defined data such as a DEM, dh/dt map/profile, SMB map/profile.
# extensions recognized are .shp, .tif, .csv
# data are assumed to be f(x,y) - if not, set ftype='z' (or anything other than xy)
# dataName is assumed to be 'z' (this is the field it will look for in .shp, .csv files)
# units are assumed to be meters (or m/yr) - if they aren't,
# be sure to specify this with unitConv flag, i.e. unitConv=365.25 for m/d
class SpatialData():

    def __init__(self, in_filename, in_dir='.', unitConv=1, ftype='xy', dataName='z'):
        FileExt = in_filename.split('.')[-1]
        if FileExt.lower() == 'tif' or FileExt.lower() == 'tiff':
            tmp = GeoImg(in_filename, in_dir=in_dir)
            ndv = tmp.gd.GetRasterBand(1).GetNoDataValue()
            X, Y = tmp.xy()
            self.x = X.reshape(-1)
            self.y = Y.reshape(-1)
            self.c, self.r = tmp.img.shape
            self.data = tmp.img.reshape(-1) * unitConv
            self.data[self.data == ndv] = np.nan
            self.img = True
        elif FileExt.lower() == 'shp':
            tmp = gpd.GeoDataFrame.from_file(in_dir + os.path.sep + in_filename)
            self.x = np.empty(0)
            self.y = np.empty(0)
            for pt in tmp['geometry']:
                self.x = np.append(self.x, pt.x)
                self.y = np.append(self.y, pt.y)
            # not sure how people would call these things
            # just assume that the default is going to be 'z'
            self.data = tmp[dataName] * unitConv
            self.img = False

        elif FileExt.lower() == 'csv':
            tmp = pd.read_csv(in_dir + os.path.sep + in_filename, sep=',|;', engine='python')
            if ftype == 'xy':
                self.x = tmp['x']
                self.y = tmp['y']
            else:
                self.x = tmp['z']
                self.y = None
            self.data = tmp[dataName] * unitConv
            self.img = False

        self.xy = ftype == 'xy'

    def interpolate(self, pts):
        if self.xy:
            interpData = griddata((self.x, self.y), self.data, (pts[0], pts[1]))
        else:
            f = interp1d(self.x, self.data)
            interpData = f(pts)
        return interpData

    def display(self, fig=None, clim=None, extent=None, cmap='RdBu'):

        if fig is None:
            fig = plt.figure()
        else:
            fig.hold(True)  # make sure we don't do away with anything

        if self.xy:
            if self.img:
                if extent is None:
                    extent = [min(self.x), max(self.x), min(self.y), max(self.y)]

                imshw = plt.imshow(self.data.reshape(self.c, self.r), extent=extent, cmap=cmap)
                if clim is not None:
                    imshw.set_clim(vmin=clim[0], vmax=clim[1])
                plt.colorbar()
                ax = fig.gca()  # get current axes
                ax.set_aspect('equal')    # set equal aspect
                ax.autoscale(tight=True)  # set axes tight

                fig.show()  # don't forget this one!
                return fig
        else:
            return


class VelData():

    def __init__(self, in_filename, in_dir='.', unitConv=1):
        FileExt = in_filename.split('.')[-1]
        if FileExt == 'shp':
            tmp = VelField(in_filename, in_dir)
            self.x = tmp.x
            self.y = tmp.y
            self.ux = tmp.ux * unitConv
            self.uy = tmp.uy * unitConv
            self.u = np.sqrt(self.ux**2 + self.uy**2)
            self.unitConv = unitConv
        elif FileExt == 'csv':
            tmp = pd.read_csv(in_dir + os.path.sep + in_filename, sep=',|;', engine='python')
            self.x = tmp['x']
            self.y = tmp['y']
            self.ux = tmp['ux'] * unitConv
            self.uy = tmp['uy'] * unitConv
            self.u = np.sqrt(self.ux**2 + self.uy**2)
            self.unitConv = unitConv

    def display(self, fig=None, cmap='spring', clim=None):

        if fig is None:
            fig = plt.figure()
        else:
            fig.hold(True)  # make sure we don't do away with anything

        qplt = plt.quiver(self.x, self.y, self.ux, self.uy, self.u, cmap=cmap)

        if clim is not None:
            qplt.set_clim(vmin=clim[0], vmax=clim[1])

        ax = fig.gca()  # get current axes
        ax.set_aspect('equal')    # set equal aspect
        ax.autoscale(tight=True)  # set axes tight

        fig.show()

        return fig

    def reverse_flow(self):
        self.ux = -self.ux
        self.uy = -self.uy


# okay, this is the big one: where we take start (end), and walk through the flowbands until we get to the end
def calculate_over_fluxgate(start, vdata, surfdata, bdot, dhdt, beddata, end=None, offset=100, gamma=0.9, plotfig=None):

    startgate = LineString(start)  # just to make sure...
    startpt = startgate.coords[0]
    flowLine1 = get_flowline(startpt, vdata, end=end)
    fluxDist = 0
    gates = []
    flowcellnumber = []
    bed = np.empty(0)

    while (startgate.length - fluxDist) > offset:
        startpt = startgate.interpolate(fluxDist+offset).coords[0]
        flowLine2 = get_flowline(startpt, vdata, end=end)

        if plotfig is not None:
            plt.plot(flowLine1.x, flowLine1.y, 'b')
            plt.plot(flowLine2.x, flowLine2.y, 'b')

        thesegates, thisbed = calculate_bed_elevs(flowLine1, flowLine2, vdata, surfdata,
                                                  bdot, dhdt, beddata, gamma, offset, plotfig)

        for j, x in enumerate(thesegates[0]):
            gates.append(Point(x, thesegates[1][j]))
            flowcellnumber.append(j)

        bed = np.append(bed, thisbed)

        flowLine1 = flowLine2
        fluxDist += offset

    return gates, bed, flowcellnumber


# this is the one where the magic happens
def calculate_bed_elevs(flowline1, flowline2, vdata, surfdata, bdot, dhdt, beddata, gamma, offset, plotfig=None):

    minlength = min(len(flowline1.x), len(flowline2.x))
    # make sure we agree on lengths of flowlines
    flow1 = FlowLine(zip(flowline1.x[0:minlength], flowline1.y[0:minlength]))
    flow2 = FlowLine(zip(flowline2.x[0:minlength], flowline2.y[0:minlength]))

    # create each cell in this flowband
    cells, widths, centers, gates = get_flow_cells(flow1, flow2, offset)

    if plotfig is not None:
        for c in cells:
            c.display(plotfig)

    # get z, dhdt, bdot for centroids
    # centersurf = griddata((surfdata.x, surfdata.y), surfdata.data, centers)
    gatesurf = surfdata.interpolate(gates)
    centersurf = surfdata.interpolate(centers)
    # if dhdt, bdot are f(z), can't just run griddata on them.
    if bdot.xy:
        centerbdot = bdot.interpolate(centers)
    else:
        centerbdot = bdot.interpolate(centersurf)
    # centerdhdt = griddata((dhdt.x, dhdt.y), dhdt.data, centers)
    # centerbdot = griddata((bdot.x, bdot.y), bdot.data, centers)
    if dhdt.xy:
        centerdhdt = dhdt.interpolate(centers)
    else:
        centerdhdt = dhdt.interpolate(centersurf)

    # need to get x,y components of velocity, and dot with the normal vector of each cell
    ux = griddata((vdata.x, vdata.y), vdata.ux, gates)
    uy = griddata((vdata.x, vdata.y), vdata.uy, gates)

    bed = np.array([beddata.interpolate((gates[0][0], gates[1][0]))])

    thickness = np.array(gatesurf[0] - bed[0])
    # the flux in is the flux at the upstream gate
    # qup = np.dot(np.array(ux[0],uy[0]), cells[0].n_up) * thickness * widths[0] * gamma
    qup = np.sqrt(ux[0]**2 + uy[0]**2) * thickness * widths[0] * gamma

    for j, c in enumerate(cells):
        # the thickness at the downstream gate is (qin - rhs)/(v_down * width_down * gamma)
        rhs = (centerbdot[j] - centerdhdt[j]) * c.area  # a somewhat crude approximation, maybe a better option?
        thickness = np.append(thickness, (qup - rhs) /
                              (np.dot(np.array(ux[j+1], uy[j+1]), c.n_dn) * widths[j+1] * gamma))
        # thickness = np.append(thickness, (qup - rhs) / (np.sqrt(ux[j+1]**2 + uy[j+1]**2) * widths[j+1] * gamma))
        bed = np.append(bed, gatesurf[j+1] - thickness[-1])  # surface at downstream boundary
        #                                                      minus calculated thicknes (at downstream bndry)

        # j+1 is the downstream index (j starts at 0, corresponds to first upstream gate)
        qup = thickness[-1] * widths[j+1] * np.dot(np.array(ux[j+1], uy[j+1]), -c.n_dn)
        # qup = thickness[-1] * widths[j+1] * np.sqrt(ux[j+1]**2 + uy[j+1]**2)

    return gates, bed


# get a flowline from the current location
def get_flowline(start, vdata, maxt=5000, end=None):
    # should figure out what the realtionship between t, ground distance is...
    t = np.arange(0, maxt / vdata.unitConv, 10 / vdata.unitConv)
    sol = odeint(velocityvect, start, t, args=(vdata,))

    this_flowline = cut_flowline(FlowLine(sol[np.isfinite(sol[:, 0]), :]), end)
    return this_flowline


# interpolate the current location to the velocity
def velocityvect(xy, t, vdata):
    uxy = np.empty(2)
    uxy[0] = griddata((vdata.x, vdata.y), vdata.ux, (xy[0], xy[1]))
    uxy[1] = griddata((vdata.x, vdata.y), vdata.uy, (xy[0], xy[1]))
    return uxy


# calculate the distance between two points
def distance(pt1, pt2):  # use shapely!
    dx = pt2[0] - pt1[0]
    dy = pt2[1] - pt1[1]
    return np.sqrt(dx*dx + dy*dy)


# returns the distance along a line described by x,y
def linedistance(x, y):
    tmp = LineString(zip(x, y))
    return [tmp.project(Point(p)) for p in zip(x, y)]


# find the width of each flowband at each vertex
def flowbandwidth(flowline1, flowline2, offsets):
    widths = np.empty(0)
    for i, pt in enumerate(flowline1.x):
        widths = np.append(widths, Point(flowline1.x[i],
                           flowline1.y[i]).distance(Point(flowline2.x[i], flowline2.y[i])))
    return widths


# get the center positions and areas for each of our cells on which to calculate continuity
# can add surf information back in later if it makes sense (?)
def get_flow_cells(flowline1, flowline2, offset):
    flowcells = []
    centerx = []
    centery = []
    widths = []
    thisdist = 0
    thisind1 = 0
    thisind2 = 0

    maxdist = min(flowline1.length, flowline2.length)
    gatex = [(flowline1.x[0]+flowline2.x[0])/2]
    gatey = [(flowline1.y[0]+flowline2.y[0])/2]

    lastpt1 = Point(flowline1.coords[0])
    lastpt2 = Point(flowline2.coords[0])
    widths.append(lastpt1.distance(lastpt2))

    while thisdist < maxdist:
        thispt1 = flowline1.interpolate(thisdist+widths[-1])
        # now, get the perp to flow at this point
        nextind1 = np.searchsorted(flowline1.linedist, thisdist+widths[-1])
        if nextind1 == 0:
            flowvect = np.array(flowline1.coords[1]) - np.array(flowline1.coords[0])
        elif nextind1 == len(flowline1.x):
            break
        else:
            flowvect = np.array(flowline1.coords[nextind1]) - np.array(flowline1.coords[nextind1-1])

        perp = norm_vector(flowvect)
        f2dist = thispt1.distance(flowline2)
        flowperp = LineString([(thispt1.x - perp[0]*1.5*f2dist, thispt1.y - perp[1]*1.5*f2dist),
                               (thispt1.x + perp[0]*1.5*f2dist, thispt1.y + perp[1]*1.5*f2dist)])
        thispt2 = flowline2.intersection(flowperp)
        if not flowline2.intersects(flowperp):
            break
        nextind2 = np.searchsorted(flowline2.linedist, flowline2.project(thispt2))

        # there might be a less convoluted way to do this...
        f1x = np.concatenate([np.asarray([lastpt1.x]), flowline1.x[thisind1:nextind1], np.asarray([thispt1.x])])
        f1y = np.concatenate([np.asarray([lastpt1.y]), flowline1.y[thisind1:nextind1], np.asarray([thispt1.y])])

        f2x = np.concatenate([np.asarray([lastpt2.x]), flowline2.x[thisind2:nextind2], np.asarray([thispt2.x])])
        f2y = np.concatenate([np.asarray([lastpt2.y]), flowline2.y[thisind2:nextind2], np.asarray([thispt2.y])])

        flow1coords = list(zip(f1x, f1y))
        flow2coords = list(zip(f2x, f2y))

        newcell = Cell(flow1coords, flow2coords)
        flowcells.append(newcell)

        centerx.append(newcell.x)
        centery.append(newcell.y)

        gatex.append(newcell.x_dn)
        gatey.append(newcell.y_dn)

        thisind1 = nextind1
        thisind2 = nextind2

        lastpt1 = thispt1
        lastpt2 = thispt2
        widths.append(lastpt1.distance(lastpt2))

        thisdist += offset

    return flowcells, widths, (centerx, centery), (gatex, gatey)


# clip flowlines to another line
def cut_flowline(thisflowline, cutline):
    if cutline is not None and thisflowline.intersects(cutline):
        IntPt = thisflowline.intersection(cutline)
        FlowDst = thisflowline.project(IntPt)
        IntInd = np.searchsorted(thisflowline.linedist, FlowDst)
        CutCoords = thisflowline.coords[0:IntInd+1]  # add one to include IntInd
        if thisflowline.linedist[IntInd] < FlowDst:
            CutCoords.append((IntPt.x, IntPt.y))
        thisflowline = FlowLine(CutCoords)
    return thisflowline


# normal vector!
def norm_vector(a):
    b = np.array([a[1], -a[0]])
    b = b / np.linalg.norm(a)
    return b



# import multiprocessing as mp
import numpy as np
from scipy.interpolate import griddata, interp1d, Rbf, RectBivariateSpline as RBS
from numpy.polynomial.polynomial import polyval, polyfit
import pybob.coreg_tools as ct
import pybob.image_tools as it
from pybob.bob_tools import bin_data
from pybob.GeoImg import GeoImg


def difference(dem1, dem2, glaciermask=None, landmask=None, outdir='.'):
    master, slave = ct.dem_coregistration(dem1, dem2, glaciermask, landmask, outdir)
    master.unmask()
    slave.unmask()

    return master.copy(new_raster=(master.img - slave.img))


def fill_holes(dDEM, method, **kwargs):
    if type(dDEM) is not GeoImg:
        raise TypeError('dDEM must be a GeoImg')
    else:
        return method(dDEM, **kwargs)


# the following (linear, elevation_mean, elevation_median, elevation_poly, neighborhood_average)
# are all methods to pass to fill_holes. E.g., filled_DEM = fill_holes(DEM, linear)
# will use linear interpolation to fill any holes in the DEM (or GeoImg raster).
# linear and neighborhood_average can be used on either a DEM or a dDEM, but elevation_* requires
# both a DEM and a dDEM to work correctly.
def linear(DEM, **kwargs):
    if type(DEM) is not GeoImg:
        raise TypeError('DEM must be a GeoImg')
    X, Y = DEM.xy()

    interp_points = DEM.img[np.isfinite(DEM.img)]
    interpX = X[np.isfinite(DEM.img)]
    interpY = Y[np.isfinite(DEM.img)]

    filled_DEM = griddata((interpX, interpY), interp_points, (X, Y))
    return DEM.copy(new_raster=filled_DEM)


def elevation_function(dDEM, **kwargs):
    if 'functype' not in kwargs:
        raise ValueError('Have to specify functype to use elevation_function.')

    return parse_elev_func_args(kwargs['functype'], dDEM, **kwargs)


def parse_elev_func_args(func_name, dDEM, **kwargs):
    # first, make sure we use one of the right function names
    if func_name not in ['mean', 'median', 'poly']:
        raise ValueError('{} not one of mean, median, poly'.format(func_name))

    if 'DEM' not in kwargs:
        raise ValueError('to use {}, you must supply a base DEM'.format(func_name))
    else:  # if we've been supplied a DEM, we have to check that it's the right size.
        DEM = kwargs['DEM']
        if type(DEM) is not GeoImg:
            raise TypeError('DEM must be a GeoImg')
        if DEM.img.shape != dDEM.img.shape:
            raise ValueError('dDEM and DEM must have the same shape')

    if 'glacier_mask' in kwargs:
        ddem_data = dDEM.img[np.logical_and(np.isfinite(dDEM.img), kwargs['glacier_mask'])]
        dem_data = DEM.img[np.logical_and(np.isfinite(dDEM.img), kwargs['glacier_mask'])]
    else:
        ddem_data = dDEM.img[np.isfinite(dDEM.img)]
        dem_data = DEM.img[np.isfinite(dDEM.img)]

    # now, figure out which of mean, median, polyfit we're using, and do it.
    if func_name == 'poly':
        if 'poly_order' not in kwargs:
            raise ValueError('poly_order must be defined to use polynomial fitting')
        # now we fit a polynomial of order poly_order to the data:
        pfit = polyfit(dem_data, ddem_data, kwargs['poly_order'])

        # need a way to put hole values in the right place
        hole_inds = np.where(np.isnan(dDEM.img))
        dem_holes = DEM.img[hole_inds]
        interp_points = polyval(dem_holes, pfit)

        dDEM.img[dem_holes] = interp_points
        return dDEM
    elif func_name == 'mean' or func_name == 'median':
        if 'bins' not in kwargs:
            # if we haven't been handed bins, we have to make our own.
            if 'bin_width' not in kwargs:
                bin_width = 50  # if we haven't been told what bin width to use, default to 50.
            else:
                bin_width = kwargs['bin_width']
            min_el = DEM.img.min() - (DEM.img.min() % bin_width)
            max_el = DEM.img.max() + (bin_width - (DEM.img.max() % bin_width))
            bins = np.arange(min_el, max_el, bin_width)
        else:
            bins = kwargs['bins']
        binned_dH = bin_data(bins, ddem_data, dem_data, mode=func_name)
        # now, we interpolate the missing DH values to the binned values.
        f_elev = interp1d(bins, binned_dH)
        # pull out missing values from dDEM
        hole_inds = np.where(np.isnan(dDEM.img))
        hole_elevs = DEM.img[hole_inds]

        filled = f_elev(hole_elevs)
        dDEM.img[hole_inds] = filled

        return dDEM
    else:
        raise ValueError('Somehow we made it this far without naming a correct fitting function. Oops.')


def neighborhood_average(dDEM, **kwargs):
    if type(dDEM) is not GeoImg:
        raise TypeError('DEM must be a GeoImg')
    if 'neighborhood_size' not in kwargs:
        raise ValueError('neighborhood_size must be defined to use ddem_tools.neighborhood_poly()')

    nradius = int(kwargs['neighborhood_size'] / dDEM.dx)

    tmp_ddem = dDEM.copy().img
    out_ddem = dDEM.copy()

    missing = np.where(np.isnan(tmp_ddem))
    missing_inds = zip(missing[0], missing[1])

    tmp_ddem[np.logical_not(kwargs['glacier_mask'])] = np.nan

    for ind in missing_inds:
        nmask = neighborhood_mask(ind, nradius, tmp_ddem)
        out_ddem.img[ind] = np.nanmean(tmp_ddem[nmask])

    return out_ddem


# kriging is like linear, can be used with a dDEM or DEM.
def kriging(DEM, **kwargs):
    pass


# RBF is simliar to linear
def radial_basis(DEM, **kwargs):
    rbfunclist = ['multiquadric', 'inverse', 'gaussian', 'linear', 'cubic', 'quintic', 'thin_plate']

    if type(DEM) is not GeoImg:
        raise TypeError('DEM must be a GeoImg')
    if 'rbfunc' not in kwargs:
        # default to linear radial basis function
        rbfunc = 'linear'
    elif kwargs['rbfunc'] not in rbfunclist:
        raise ValueError('unknown radial basis function {} requested. \
                          Check scipy.interpolate.Rbf docs for details'.format(kwargs['rbfunc']))
    else:
        rbfunc = kwargs['rbfunc']

    X, Y = DEM.xy()
    vals = ~np.isnan(DEM.img)
    f = Rbf(X[vals], Y[vals], DEM.img[vals], function=rbfunc)
    out_dem = f(X, Y)

    return DEM.copy(new_raster=out_dem)


# RBS
def RectBivariateSpline(DEM, **kwargs):
    if type(DEM) is not GeoImg:
        raise TypeError('DEM must be a GeoImg')
    X, Y = DEM.xy()
    f = RBS(Y[0:, 0], X[0, 0:], DEM.img)
    out_dem = f.ev(Y, X)  # might have to flip upside-down?

    return DEM.copy(new_raster=out_dem)


# some helper functions
def neighborhood_mask(centerind, radius, array):
    a, b = centerind
    nx, ny = array.shape
    y, x = np.ogrid[-a:nx-a, -b:ny-b]
    mask = x*x + y*y <= radius*radius

    return mask


# interpolate on a glacier-by-glacier basis, rather than on the DEM as a whole.
def fill_holes_individually(dDEM, glacshapes, functype, burn_handle=None, **kwargs):
    # first, get the individual glacier mask.
    ind_glac_mask, ind_glac_vals = it.rasterize_polygons(dDEM, glacshapes, burn_handle)

    filled_ddem = dDEM.copy()

    # if we already have a glacier mask defined, remove it.
    if 'glacier_mask' in kwargs:
        del kwargs['glacier_mask']

    for glac in ind_glac_vals:
        tmp_mask = ind_glac_mask == glac
        tmp_dem = fill_holes(dDEM, elevation_function, glacier_mask=tmp_mask, functype=functype, **kwargs)

        filled_ddem.img[tmp_mask] = tmp_dem.img[tmp_mask]
    return filled_ddem


# normalize glacier elevations given a shapefile of glacier elevations
def normalize_glacier_elevations(DEM, glacshapes, burn_handle=None):
    ind_glac_mask, ind_glac_vals = it.rasterize_polygons(DEM, glacshapes, burn_handle)
    normed_els = DEM.copy().img

    for i in ind_glac_vals:
        glac_inds = np.where(ind_glac_mask == i)
        glac_els = DEM.img[glac_inds]
        if glac_els.size > 0:
            max_el = np.nanmax(glac_els)
            min_el = np.nanmin(glac_els)
            normed_els[glac_inds] = (glac_els - min_el) / (max_el - min_el)

    normDEM = DEM.copy(new_raster=normed_els)

    return normDEM, ind_glac_mask, ind_glac_vals


# now we actually sum up the ddem values that fall within each glacier
# input from a shapefile
def calculate_volume_changes(dDEM, glacier_shapes, burn_handle=None):
    ind_glac_mask, ind_glac_vals = it.rasterize_polygons(dDEM, glacier_shapes, burn_handle=burn_handle)
    ind_vol_chgs = np.zeros(ind_glac_vals.shape)

    for i, ind in ind_glac_vals:
        glac_inds = np.where(ind_glac_mask == ind)
        glac_chgs = dDEM.img[glac_inds]
        # get the volume change by summing dh/dt, multiplying by cell
        ind_vol_chgs[i] = np.nansum(glac_chgs) * np.abs(dDEM.dx) * np.abs(dDEM.dy)

    return ind_glac_vals, ind_vol_chgs


# create an area-altitude distribution for a DEM and a glacier shapefile
def area_alt_dist(DEM, glacier_shapes, glacier_inds=None, bin_width=50):
    if type(DEM) is not GeoImg:
        raise TypeError('DEM must be a GeoImg')
    if glacier_inds is None:
        dem_data = DEM.img[glacier_shapes]

        min_el = DEM.img.min() - (DEM.img.min() % bin_width)
        max_el = DEM.img.max() + (bin_width - (DEM.img.max() % bin_width))
        bins = np.arange(min_el, max_el, bin_width)
        aads, _ = np.histogram(dem_data, bins=bins)
        aads = aads * np.abs(DEM.dx) * np.abs(DEM.dy)
    else:
        bins = []
        aads = []

        for i in glacier_inds:
            dem_data = DEM.img[glacier_shapes == i]
            min_el = dem_data.min() - (dem_data.min() % bin_width)
            max_el = dem_data.max() + (bin_width - (dem_data.max() % bin_width))
            thisbin = np.arange(min_el, max_el, bin_width)
            thisaad, _ = np.histogram(dem_data, bins=thisbin)

            bins.append(thisbin)
            aads.append(thisaad * np.abs(DEM.dx) * np.abs(DEM.dy))  # make it an area!

    return bins, aads



#!/usr/bin/env python
import glob, pandas as pd, argparse

def main():
	# set up command line arguments
	parser = argparse.ArgumentParser(description="Read a folder of browse images downloaded from NASA reverb, return the granule name for each scene.", formatter_class=argparse.RawDescriptionHelpFormatter)

	parser.add_argument('csvfile', action='store', type=str, default='.', help='name of reverb csv file to read in')
	parser.add_argument('--gid', action='store_true', default=False, help='print Granule ID instead of Granule UR [False].')
	args = parser.parse_args()

	mydata = pd.read_csv(args.csvfile)

	browselist = glob.glob('*.jpg')

	for img in browselist:
		if args.gid:
			print mydata[mydata['Browse URLs'].str.contains(img)]['Producer Granule ID'].values[0]
		else:
			print mydata[mydata['Browse URLs'].str.contains(img)]['Granule UR'].values[0]			

if __name__ == "__main__":
    main()



import os
import numpy as np
import geopandas as gpd
import matplotlib.pyplot as plt


class VelField(object):

    def __init__(self, in_filename, in_dir='.'):
        self.filename = in_filename
        self.in_dir_path = in_dir
        self.in_dir_abs_path = os.path.abspath(in_dir)

        infile = gpd.GeoDataFrame.from_file(in_dir + os.path.sep + in_filename)

        self.x = np.empty(0)
        self.y = np.empty(0)

        for pt in infile['geometry']:
            self.x = np.append(self.x, pt.x)
            self.y = np.append(self.y, pt.y)

        self.ux = infile['ux'].as_matrix()
        self.uy = infile['uy'].as_matrix()
        self.corr_max = None
        self.dcorr = None
        self.z = None
        # self.date = infile['centerdate'].as_matrix()
        if 'corr_max' in infile.keys():
            self.corr_max = infile['corr_max'].as_matrix()
        if 'dcorr' in infile.keys():
            self.dcorr = infile['dcorr'].as_matrix()
        if 'z' in infile.keys():
            self.z = infile['z'].as_matrix()
        # self.img1 = infile['img1'].as_matrix()

    def display_vector_field(self, fig=None, cmap='spring', clim=None):

        if fig is None:
            fig = plt.figure()
        else:
            fig.hold(True)  # make sure we don't do away with anything

        vv = np.sqrt(self.ux**2 + self.uy**2)
        qplt = plt.quiver(self.x, self.y, self.ux, self.uy, vv, cmap=cmap)

        if clim is not None:
            qplt.set_clim(vmin=clim[0], vmax=clim[1])

        ax = fig.gca()  # get current axes
        ax.set_aspect('equal')  # set equal aspect
        ax.autoscale(tight=True)  # set axes tight

        fig.show()

        return fig



import os
import numpy as np
import cv2
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from matplotlib.pyplot import savefig
from mpl_toolkits.axes_grid1 import make_axes_locatable


def save_results_quicklook(img, raster, com_ext, outfilename, vmin=0, vmax=10, sfact=2, output_directory='.'):
    xmin, ymin, xmax, ymax = com_ext
    ext = [xmin, xmax, ymin, ymax]

    fig = img.overlay(raster, extent=ext, sfact=sfact, vmin=vmin, vmax=vmax, showfig=False)

    savefig(os.path.join(output_directory, outfilename + '.png'), bbox_inches='tight', dpi=200)
    return fig


def plot_geoimg_sidebyside(img1, img2, com_extent=None, fig=None, cmap='gray', output_directory='.', filename=None):
    if fig is None:
        fig = plt.figure(facecolor='w', figsize=(16.5, 16.5), dpi=200)
    elif isinstance(fig, Figure):
        fig
    else:
        fig = plt.figure(fig)

    ax1 = plt.subplot(121, axisbg=(0.1, 0.15, 0.15))
    if com_extent is not None:
        plt.imshow(img1, interpolation='nearest', cmap=cmap, extent=com_extent)
    else:
        plt.imshow(img2, interpolation='nearest', cmap=cmap)

    ax2 = plt.subplot(122, axisbg=(0.1, 0.15, 0.15))
    if com_extent is not None:
        plt.imshow(img2, interpolation='nearest', cmap=cmap, extent=com_extent)
    else:
        plt.imshow(img2, interpolation='nearest', cmap=cmap)

    if filename is not None:
        savefig(os.path.join(output_directory, filename), bbox_inches='tight', dpi=200)

    return fig, ax1, ax2


def plot_nice_histogram(data_values, outfilename, output_directory):
    fig = plt.figure(facecolor='w', figsize=(16.5, 16.5), dpi=200)

    outputs = plt.hist(data_values, 25, alpha=0.5, normed=1)

    ylim = plt.gca().get_ylim()
    mu = data_values.mean()
    sigma = data_values.std()

    plt.plot([mu, mu], ylim, 'k--')
    plt.plot([mu+sigma, mu+sigma], ylim, 'r--')
    plt.plot([mu+2*sigma, mu+2*sigma], ylim, 'r--')

    plt.text(mu, np.mean(ylim), r'$\mu$', fontsize=24, color='k')
    plt.text(mu+sigma, np.mean(ylim), r'$\mu + \sigma$', fontsize=24, color='r')
    plt.text(mu+2*sigma, np.mean(ylim), r'$\mu + 2\sigma$', fontsize=24, color='r')
    # just to be sure
    plt.gca().set_ylim(ylim)

    savefig(output_directory + os.path.sep + outfilename + '_vdist.png', bbox_inches='tight', dpi=200)
    return fig, outputs


def plot_chips_corr_matrix(srcimg, destimg, corrmat):
    fig = plt.figure()
    plt.ion()

    ax1 = plt.subplot(131)
    plt.imshow(srcimg, cmap='gray')
    ax1.set_title('source image')

    ax2 = plt.subplot(132)
    plt.imshow(destimg, cmap='gray')
    ax2.set_title('dest. image')

    peak1_loc = cv2.minMaxLoc(corrmat)[3]

    ax3 = plt.subplot(133)
    plt.imshow(corrmat, cmap='jet')
    plt.plot(peak1_loc[0], peak1_loc[1], 'w+')
    ax3.set_title('correlation')

    plt.show()

    return fig


def plot_ddem_results(img, colormap='seismic', caxsize="2.5%", clim=None):
    fig = img.display(cmap=colormap)

    if clim is not None:
        plt.clim(clim[0], clim[1])

    ax = plt.gca()

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size=caxsize, pad=0.1)
    plt.colorbar(cax=cax)

    return fig, ax, cax



#!/usr/bin/env python
import gdal, sys, os, argparse, numpy as np
from pybob.GeoImg import GeoImg
from pybob.image_tools import generate_panchrome

def main():
    parser = argparse.ArgumentParser(description = "Generate simulated panchromatic image for Landsat TM data.")
    parser.add_argument("inputscene", type=str, help="Base Landsat scene name (do not specify bands) to be read in.")
    parser.add_argument("-o", "--outputscene", type=str, help="Output scene name (if unspecified, defaults to inputscenename_B8.TIF)")
    args = parser.parse_args()

    if args.outputscene is None:
	    args.outputscene = args.inputscene

    outfilename = args.outputscene + "_B8.TIF"

    # first, read in the bands (4, 3, 2)
    B4 = GeoImg( args.inputscene + "_B4.TIF" )
    B3 = GeoImg( args.inputscene + "_B3.TIF" )
    B2 = GeoImg( args.inputscene + "_B2.TIF" )

    # now, make a new band
    B8sim = 0.5 * B4.img + 0.2 * B3.img + 0.3 * B2.img

    B8 = B4.copy(new_raster=B8sim)

    B8.write(outfilename)

if __name__ == "__main__":
    main()



import gdal
import os
import ogr
import numpy as np
import multiprocessing as mp
import scipy.ndimage as ndimage
import scipy.ndimage.filters as filters
from skimage.feature import greycomatrix, greycoprops
from pybob.bob_tools import parse_lsat_scene, round_down
from pybob.GeoImg import GeoImg


def generate_panchrome(imgname, outname=None, out_dir='.', interactive=False):
    if outname is None:
        outname = imgname + '_B8.TIF'

    B5 = GeoImg(imgname + '_B5.TIF')
    B4 = GeoImg(imgname + '_B4.TIF')
    B3 = GeoImg(imgname + '_B3.TIF')
    B2 = GeoImg(imgname + '_B2.TIF')

    B8sim = 0.45 * B4.img + 0.2 * B3.img + 0.25 * B2.img + 0.1 * B5.img
    B8 = B4.copy(new_raster=B8sim)
    B8.write(outname, out_folder=out_dir)

    if interactive:
        return B8


def composite_raster(band1name, band2name, band3name, outname, out_dir='.', in_dir='.', driver='GTiff'):
    band1 = GeoImg(band1name, in_dir=in_dir)
    band2 = GeoImg(band2name, in_dir=in_dir)
    band3 = GeoImg(band3name, in_dir=in_dir)

    driver = gdal.GetDriverByName(driver)

    ncols = band1.npix_x
    nrows = band1.npix_y
    nband = 3
    datatype = band1.gd.GetRasterBand(1).DataType

    out = driver.Create(out_dir + os.path.sep + outname, ncols, nrows, nband, datatype)

    out.SetGeoTransform(band1.gt)
    out.SetProjection(band1.proj)

    out.GetRasterBand(1).WriteArray(band1.gd.ReadAsArray())
    out.GetRasterBand(2).WriteArray(band2.gd.ReadAsArray())
    out.GetRasterBand(3).WriteArray(band3.gd.ReadAsArray())

    for i in range(3):
        out.GetRasterBand(i+1).FlushCache()


def write_landsat_rgb(scenename, outname, out_dir='.', in_dir='.', driver='GTiff'):
    # takes the scenename, finds bands 1-3 (tm, etm), or 2-4 (oli), then writes outfilename.TIF, the RGB composite.
    sensor = parse_lsat_scene(scenename)[0]

    if (sensor == 'LC8' or sensor == 'LO8'):
        band1name = scenename + '_B4.TIF'
        band2name = scenename + '_B3.TIF'
        band3name = scenename + '_B2.TIF'
    else:
        band1name = scenename + '_B3.TIF'
        band2name = scenename + '_B2.TIF'
        band3name = scenename + '_B1.TIF'

    composite_raster(band1name, band2name, band3name, outname, out_dir, in_dir, driver)


def find_peaks(image, neighborhood_size=5, threshold=1500):
    data_max = filters.maximum_filter(image, neighborhood_size)
    maxima = (image == data_max)
    data_min = filters.minimum_filter(image, neighborhood_size)
    diff = ((data_max - data_min) > threshold)
    maxima[diff == 0] = 0

    labeled, num_objects = ndimage.label(maxima)
    ij = np.array(ndimage.center_of_mass(image, labeled, range(1, num_objects+1)))

    return ij


def create_mask_from_shapefile(geoimg, shapefile):
    # load shapefile, rasterize to raster's extent
    masksource = ogr.Open(shapefile)
    masklayer = masksource.GetLayer()
    # masksrs = masklayer.GetSpatialRef()

    masktarget = gdal.GetDriverByName('MEM').Create('', geoimg.npix_x, geoimg.npix_y, 1, gdal.GDT_Byte)
    masktarget.SetGeoTransform((geoimg.xmin, geoimg.dx, 0, geoimg.ymax, 0, geoimg.dy))
    masktarget.SetProjection(geoimg.proj)
    masktarget.GetRasterBand(1).Fill(0)
    gdal.RasterizeLayer(masktarget, [1], masklayer)
    mask = masktarget.GetRasterBand(1).ReadAsArray()
    mask[mask != 0] = 1

    return mask == 1


def rasterize_polygons(geoimg, shapefile, burn_handle=None, dtype=gdal.GDT_Int16):
    polysource = ogr.Open(shapefile, 0)  # read-only
    polylayer = polysource.GetLayer()

    if burn_handle is None:
        if polylayer.GetFIDColumn() == '':
            polylayer = add_fid_column(polylayer)
            burn_handle = 'TmpFID'
        else:
            burn_handle = polylayer.GetFIDColumn()

    target = gdal.GetDriverByName('MEM').Create('', geoimg.npix_x, geoimg.npix_y, 1, dtype)
    target.SetGeoTransform((geoimg.xmin, geoimg.dx, 0, geoimg.ymax, 0, geoimg.dy))
    target.SetProjection(geoimg.proj)
    target.GetRasterBand(1).Fill(-1)

    gdal.RasterizeLayer(target, [1], polylayer, options=["Attribute={}".format(burn_handle)])
    thisrast = target.GetRasterBand(1).ReadAsArray()

    thesevals = [feat.GetField(burn_handle) for feat in polylayer]

    return thisrast, np.array(thesevals)


def add_fid_column(layer):
    tmpdrv = ogr.GetDriverByName('MEMORY')
    tmpsrc = tmpdrv.CreateDataSource('memData')
    tmpdrv.Open('memData', 1)
    newlayer = tmpsrc.CopyLayer(layer, '', ['OVERWRITE=YES'])

    fid_column = ogr.FieldDefn('TmpFID', ogr.OFTInteger)
    newlayer.CreateField(fid_column)

    for feat in newlayer:
        feat.SetField('TmpFID', feat.GetFID())

    return newlayer


def get_common_xy(bbox, img1, img2):

    if (bbox is None):
        #     find common x,y (using original UTM or other geo cartesian coordinates) box between images
        max_x = min([img1.xmax, img2.xmax])
        max_y = min([img1.ymax, img2.ymax])
        min_x = max([img1.xmin, img2.xmin])
        min_y = max([img1.ymin, img2.ymin])
    else:  # read area from command line (still UTM etc.)
        max_x = bbox[2]
        max_y = bbox[3]
        min_x = bbox[0]
        min_y = bbox[1]

    return max_x, max_y, min_x, min_y


def calc_glcm_params(img):
    glcmcontr = np.zeros(img.shape)
    glcmdissim = np.zeros(img.shape)
    glcmhomog = np.zeros(img.shape)
    glcmenergy = np.zeros(img.shape)
    glcmcorrel = np.zeros(img.shape)
    glcmASM = np.zeros(img.shape)

    # have to make sure that the raster is uint16 or uint8
    tmpimg = img.astype(np.uint8)

    for i in range(img.shape[0]):
        for j in range(img.shape[1]):
            if (i < 3) or (j < 3):
                continue
            if (i > (img.shape[0] - 4)) or (j > (img.shape[1] - 4)):
                continue
            glcm_window = tmpimg[i-3:i+4, j-3:j+4]
            glcm = greycomatrix(glcm_window, [1], [0], symmetric=True, normed=True)

            glcmcontr[i, j] = greycoprops(glcm, 'contrast')
            glcmdissim[i, j] = greycoprops(glcm, 'dissimilarity')
            glcmhomog[i, j] = greycoprops(glcm, 'homogeneity')
            glcmenergy[i, j] = greycoprops(glcm, 'energy')
            glcmcorrel[i, j] = greycoprops(glcm, 'correlation')
            glcmASM[i, j] = greycoprops(glcm, 'ASM')
            glcm = None
            glcm_window = None

    return glcmcontr, glcmdissim, glcmhomog, glcmenergy, glcmcorrel, glcmASM


def get_nicely_shaped_img(img, overlap):
    # get the new shape of the image
    new_cols = round_down(img.img.shape[1], overlap)
    new_rows = round_down(img.img.shape[0], overlap)

    new_img = img.img[0:new_rows, 0:new_cols]
    new_exts = img.ij2xy((new_rows, new_cols))

    if img.dy < 0:
        new_ext = [img.xmin, new_exts[0], new_exts[1], img.ymax]
    else:
        new_ext = [img.xmin, new_exts[0], img.ymin, new_exts[1]]

    return img.copy(new_raster=new_img, new_extent=new_ext)


def splitter(img, nblocks, overlap=0):
    blocks = []

    if np.array(nblocks).size == 1:
        nblocks = np.array([nblocks, nblocks])
    new_width = int(img.shape[1] / nblocks[1])
    new_height = int(img.shape[0] / nblocks[0])

    for j in range(nblocks[1]):
        for i in range(nblocks[0]):
            lind = max(0, j*new_width - overlap)
            rind = min(img.shape[1], (j+1)*new_width + overlap)

            tind = max(0, i*new_height - overlap)
            bind = min(img.shape[0], (i+1)*new_height + overlap)

            blocks.append(img[tind:bind, lind:rind])

    return blocks


def stitcher(outputs, nblocks, overlap=0):
    stitched_arrays = []  # create an empty list to put our stitched results in
    outarrays = []  # same

    # check if nblocks is a scalar or not.
    if np.array(nblocks).size == 1:
        nblocks = np.array([nblocks, nblocks])

    for out in range(len(outputs[0])):
        stitched_arrays.append([])
        outarrays.append([])
        for j in range(nblocks[0]):
            stitched_arrays[out].append(outputs[j*nblocks[0]][out][overlap:-overlap, overlap:-overlap])
            for i in range(1, nblocks[1]):
                outind = i + j*nblocks[0]
                stitched_arrays[out][j] = np.vstack((stitched_arrays[out][j],
                                                     outputs[outind][out][overlap:-overlap, overlap:-overlap]))
        stitched_arrays[out] = np.hstack(stitched_arrays[out])
        outarrays[out] = np.zeros((stitched_arrays[out].shape[0]+2*overlap, stitched_arrays[out].shape[1]+2*overlap))
        outarrays[out][overlap:-overlap, overlap:-overlap] = stitched_arrays[out]

    return outarrays


def parallel_img_proc(fun, funargs, nproc=2):
    if nproc > mp.cpu_count():
        print "{} cores specified, but I can only find \
               {} cores on this machine. I'll use those.".format(nproc, mp.cpu_count())

    pool = mp.Pool(nproc)
    outputs = pool.map(fun, funargs)
    pool.close()

    return outputs



from setuptools import setup

setup(name='pybob',
      version='0.11.5-1',
      description='Collection of geospatial and other tools I find useful.',
      url='http://github.com/iamdonovan/pybob',
      author='Bob McNabb',
      author_email='robertmcnabb@gmail.com',
      license='MIT',
      packages=['pybob'],
      install_requires=[
                        'numpy', 'scipy', 'matplotlib', 'fiona',
                        'shapely', 'opencv-python', 'pandas', 'geopandas',
                        'scikit-image',
                         ],
      scripts=['bin/dem_coregistration.py', 'bin/generate_panchromatic.py',
               'bin/print_reverb_browse_urls.py', 'bin/print_reverb_granule_names.py'],
      zip_safe=False)



import os
import random
import datetime as dt
import numpy as np
import matplotlib.pyplot as plt
import gdal
import osr
from mpl_toolkits.axes_grid1 import make_axes_locatable
from pybob.bob_tools import standard_landsat
from scipy.interpolate import griddata


def get_file_info(in_filestring):
    in_filename = os.path.basename(in_filestring)
    in_dir = os.path.dirname(in_filestring)
    if in_dir == '':
        in_dir = '.'
    return in_filename, in_dir


class GeoImg(object):

    def __init__(self, in_filename, in_dir=None, datestr=None, datefmt='%m/%d/%y'):
        # if in_dir is None, we have to figure it out from the in_filename
        # otherwise, we load normally.
        if type(in_filename) is gdal.Dataset:
            self.filename = None
            self.in_dir_path = None
            self.in_dir_abs_path = None
            self.gd = in_filename
        elif type(in_filename) is str:
            if in_dir is None:
                in_filename, in_dir = get_file_info(in_filename)
            self.filename = in_filename
            self.in_dir_path = in_dir
            self.in_dir_abs_path = os.path.abspath(in_dir)
            self.gd = gdal.Open(os.path.join(self.in_dir_path, self.filename))
        else:
            raise Exception('in_filename must be a string or a gdal Dataset')

        self.gt = self.gd.GetGeoTransform()
        self.proj = self.gd.GetProjection()
        self.epsg = int(''.join(filter(lambda x: x.isdigit(), self.proj.split(',')[-1])))
        self.spatialReference = osr.SpatialReference()
        dump = self.spatialReference.ImportFromEPSG(self.epsg)  # do this to make our SRS nice and easy for osr later.
        del dump
        self.intype = self.gd.GetDriver().ShortName
        self.npix_x = self.gd.RasterXSize
        self.npix_y = self.gd.RasterYSize
        self.xmin = self.gt[0]
        self.xmax = self.gt[0] + self.npix_x * self.gt[1] + self.npix_y * self.gt[2]
        self.ymin = self.gt[3] + self.npix_x * self.gt[4] + self.npix_y * self.gt[5]
        self.ymax = self.gt[3]
        self.dx = self.gt[1]
        self.dy = self.gt[5]
        self.UTMtfm = [self.xmin, self.ymax, self.dx, self.dy]
        self.NDV = self.gd.GetRasterBand(1).GetNoDataValue()
        self.img = self.gd.ReadAsArray().astype(np.float32)
        self.dtype = self.gd.ReadAsArray().dtype
        if self.NDV is not None:
            self.img[self.img == self.NDV] = np.nan

        if self.filename is not None:
            if (datestr is not None):
                self.imagedatetime = dt.datetime.strptime(datestr, datefmt)
            elif (self.filename[0] == 'L'):  # if it looks like a Landsat file
                self.filename = standard_landsat(self.filename)
                self.sensor = self.filename[0:3]
                self.path = int(self.filename[3:6])
                self.row = int(self.filename[6:9])
                self.year = int(self.filename[9:13])
                self.doy = int(self.filename[13:16])
                self.imagedatetime = dt.date.fromordinal(dt.date(self.year-1, 12, 31).toordinal()+self.doy)
            else:
                self.datetime = None
        else:
            self. datetime = None

        self.img_ov2 = self.img[0::2, 0::2]
        self.img_ov10 = self.img[0::10, 0::10]

    def info(self):
        print 'Driver:             {}'.format(self.gd.GetDriver().LongName)
        if self.intype != 'MEM':
            print 'File:               {}'.format(self.in_dir_path + os.path.sep + self.filename)
        else:
            print 'File:               {}'.format('in memory')
        print 'Size:               {}, {}'.format(self.npix_x, self.npix_y)
        print 'Coordinate System:  EPSG:{}'.format(self.epsg)
        print 'NoData Value:       {}'.format(self.NDV)
        print 'Pixel Size:         {}, {}'.format(self.dx, self.dy)
        print 'Upper Left Corner:  {}, {}'.format(self.xmin, self.ymax)
        print 'Lower Right Corner: {}, {}'.format(self.xmax, self.ymin)
        print '[MAXIMUM]:          {}'.format(np.nanmax(self.img))
        print '[MINIMUM]:          {}'.format(np.nanmin(self.img))
        # print '[MEAN]:             {}'.format(np.nanmean(self.img))
        # print '[MEDIAN]:           {}'.format(np.nanmedian(self.img))

    def display(self, fig=None, cmap='gray', extent=None, sfact=None, showfig=True, band=[0, 1, 2]):
        if fig is None:
            fig = plt.figure(facecolor='w')
            fig.hold(True)
        else:
            fig.hold(True)

        if extent is None:
            extent = [self.xmin, self.xmax, self.ymin, self.ymax]

            mini = 0
            maxi = self.npix_y
            minj = 0
            maxj = self.npix_x
        else:
            xmin, xmax, ymin, ymax = extent
            mini, minj = self.xy2ij((xmin, ymax))
            maxi, maxj = self.xy2ij((xmax, ymin))

            mini += 0.5
            minj += 0.5

            maxi -= 0.5
            maxj -= 0.5  # subtract the .5 for half a pixel, add 1 for slice

        # if we only have one band, plot it.
        if self.gd.RasterCount == 1:
            if sfact is None:
                showimg = self.img[int(mini):int(maxi+1), int(minj):int(maxj+1)]
            else:
                showimg = self.img[int(mini):int(maxi+1):sfact, int(minj):int(maxj+1):sfact]
            plt.imshow(showimg, extent=extent, cmap=cmap)
        elif type(band) is int:
            if sfact is None:
                showimg = self.img[band][int(mini):int(maxi+1), int(minj):int(maxj+1)]
            else:
                showimg = self.img[band][int(mini):int(maxi+1):sfact, int(minj):int(maxj+1):sfact]
            plt.imshow(showimg, extent=extent, cmap=cmap)
        else:  # if we have more than one band and we've asked to display them all, do it.
            if sfact is None:
                b1 = self.img[band[0]][int(mini):int(maxi+1), int(minj):int(maxj+1)]
                b2 = self.img[band[1]][int(mini):int(maxi+1), int(minj):int(maxj+1)]
                b3 = self.img[band[2]][int(mini):int(maxi+1), int(minj):int(maxj+1)]
            else:
                b1 = self.img[band[0]][int(mini):int(maxi+1):sfact, int(minj):int(maxj+1):sfact]
                b2 = self.img[band[1]][int(mini):int(maxi+1):sfact, int(minj):int(maxj+1):sfact]
                b3 = self.img[band[2]][int(mini):int(maxi+1):sfact, int(minj):int(maxj+1):sfact]
            rgb = np.dstack([b1, b2, b3]).astype(self.dtype)
            plt.imshow(rgb, extent=extent)

        ax = fig.gca()  # get current axes
        ax.set_aspect('equal')    # set equal aspect
        ax.autoscale(tight=True)  # set axes tight

        if showfig:
            fig.show()  # don't forget this one!

        return fig

    def write(self, outfilename, out_folder='.', driver='GTiff', datatype=gdal.GDT_Float32):
        driver = gdal.GetDriverByName(driver)

        ncols = self.npix_x
        nrows = self.npix_y
        nband = 1

        out = driver.Create(out_folder + os.path.sep + outfilename, ncols, nrows, nband, datatype)

        setgeo = out.SetGeoTransform(self.gt)
        setproj = out.SetProjection(self.proj)
        if self.NDV is not None:
            self.img[np.isnan(self.img)] = self.NDV

        write = out.GetRasterBand(1).WriteArray(self.img)
        if self.NDV is not None:
            out.GetRasterBand(1).SetNoDataValue(self.NDV)

        out.GetRasterBand(1).FlushCache()
        del setgeo, setproj, write

    def copy(self, new_raster=None, new_extent=None, driver='MEM', filename='', newproj=None):
        drv = gdal.GetDriverByName(driver)
        if driver == 'MEM':
            filename = ''
        elif driver == 'GTiff' and filename == '':
            raise Exception('must specify an output filename')

        if (new_raster is None and new_extent is None):
            npix_y, npix_x = self.img.shape
            new_raster = self.img  # give the same raster
            newgt = self.gt
        elif (new_raster is not None and new_extent is None):
            # copy the geoimg and replace the raster with new_raster
            npix_y, npix_x = np.array(new_raster).shape

            dx = (self.xmax - self.xmin) / float(npix_x)
            dy = (self.ymin - self.ymax) / float(npix_y)
            newgt = (self.xmin, dx, 0, self.ymax, 0, dy)
        elif (new_raster is not None and new_extent is not None):
            # copy the geoimg and replace the raster with new_raster
            npix_y, npix_x = np.array(new_raster).shape
            # new_extent should be xmin, xmax, ymin, ymax
            dx = (new_extent[1] - new_extent[0]) / float(npix_x)
            dy = (new_extent[2] - new_extent[3]) / float(npix_y)
            newgt = (new_extent[0], dx, 0, new_extent[3], 0, dy)
        else:
            raise Exception('If new extent is specified, you must also specify the new raster to be used!')
        newGdal = drv.Create(filename, npix_x, npix_y, 1, gdal.GDT_Float32)
        wa = newGdal.GetRasterBand(1).WriteArray(new_raster)
        sg = newGdal.SetGeoTransform(newgt)

        if newproj is None:
            newproj = self.proj

        sp = newGdal.SetProjection(newproj)

        if self.NDV is not None:
            newGdal.GetRasterBand(1).SetNoDataValue(self.NDV)

        del wa, sg, sp

        return GeoImg(newGdal)

    # return X,Y grids of coordinates for each pixel
    def xy(self, ctype='corner'):
        xx = np.linspace(self.xmin, self.xmax, self.npix_x+1)

        if self.dy < 0:
            yy = np.linspace(self.ymax, self.ymin, self.npix_y+1)
        else:
            yy = np.linspace(self.ymin, self.ymax, self.npix_y+1)

        if ctype == 'center':
            xx += self.dx / 2  # shift by half a pixel
            yy += self.dy / 2

        return np.meshgrid(xx[:-1], yy[:-1])  # drop the last element

    def reproject(self, dst_raster, driver='MEM', filename=''):

        drv = gdal.GetDriverByName(driver)
        if driver == 'MEM':
            filename = ''
        elif driver == 'GTiff' and filename == '':
            raise Exception('must specify an output filename')

        dest = drv.Create('', dst_raster.npix_x, dst_raster.npix_y, 1, gdal.GDT_Float32)
        dest.SetProjection(dst_raster.proj)
        dest.SetGeoTransform(dst_raster.gt)
        gdal.ReprojectImage(self.gd, dest, self.proj, dst_raster.proj, gdal.GRA_Bilinear)

        if dst_raster.NDV is not None:
            dest.GetRasterBand(1).SetNoDataValue(dst_raster.NDV)
        out = GeoImg(dest)
        out.img[np.isnan(dst_raster.img)] = np.nan

        return out

    def shift(self, xshift, yshift):
        gtl = list(self.gt)
        gtl[0] += xshift
        gtl[3] += yshift
        self.gt = tuple(gtl)
        self.gd.SetGeoTransform(self.gt)
        self.xmin = self.gt[0]
        self.xmax = self.gt[0] + self.npix_x * self.gt[1] + self.npix_y * self.gt[2]
        self.ymin = self.gt[3] + self.npix_x * self.gt[4] + self.npix_y * self.gt[5]
        self.ymax = self.gt[3]

    def ij2xy(self, ij):
        x = self.UTMtfm[0]+((ij[1]+0.5)*self.UTMtfm[2])
        y = self.UTMtfm[1]+((ij[0]+0.5)*self.UTMtfm[3])

        return x, y

    def xy2ij(self, xy):
        x = xy[0]
        y = xy[1]
        j = int((x-self.UTMtfm[0])/self.UTMtfm[2]) - 0.5  # if python started at 1, + 0.5
        i = int((y-self.UTMtfm[1])/self.UTMtfm[3]) - 0.5  # if python started at 1, + 0.5

        return i, j

    def set_NDV(self, NDV):
        self.NDV = NDV
        self.gd.GetRasterBand(1).SetNoDataValue(NDV)
        self.img[self.img == self.NDV] = np.nan

    def subimages(self, N, Ny=None, sBuffer=0):
        Nx = N
        if Ny is None:
            Ny = N

        new_width = int(np.floor(self.npix_x / float(Nx)))
        new_height = int(np.floor(self.npix_y / float(Ny)))

        simages = []
        for j in range(Nx):
            for i in range(Ny):
                lind = max(0, j*new_width-sBuffer)
                rind = min(self.npix_x, (j+1)*new_width + sBuffer)
                tind = max(0, i*new_height-sBuffer)
                bind = min(self.npix_y, (i+1)*new_height + sBuffer)

                imgN = self.img[tind:bind, lind:rind]
                xmin, ymin = self.ij2xy((bind, lind))
                xmax, ymax = self.ij2xy((tind, rind))
                extN = [xmin, xmax, ymin, ymax]
                newGimg = self.copy(new_raster=imgN, new_extent=extN)
                simages.append(newGimg)

        return simages

    def crop_to_extent(self, extent):
        if isinstance(extent, plt.Figure):
            xmin, xmax = extent.gca().get_xlim()
            ymin, ymax = extent.gca().get_ylim()
        else:
            xmin = extent[0]
            xmax = extent[1]
            ymin = extent[2]
            ymax = extent[3]

        npix_x = int(np.round((xmax - xmin) / float(self.dx)))
        npix_y = int(np.round((ymin - ymax) / float(self.dy)))

        dx = (xmax - xmin) / float(npix_x)
        dy = (ymin - ymax) / float(npix_y)

        drv = gdal.GetDriverByName('MEM')
        dest = drv.Create('', npix_x, npix_y, 1, gdal.GDT_Float32)
        dest.SetProjection(self.proj)
        newgt = (xmin, dx, 0.0, ymax, 0.0, dy)
        dest.SetGeoTransform(newgt)
        gdal.ReprojectImage(self.gd, dest, self.proj, self.proj, gdal.GRA_Bilinear)

        if self.NDV is not None:
            dest.GetRasterBand(1).SetNoDataValue(self.NDV)

        return GeoImg(dest)

    def overlay(self, raster, extent=None, vmin=0, vmax=10, sfact=None, showfig=True):
        fig = self.display(extent=extent, sfact=sfact, showfig=showfig)

        if showfig:
            plt.ion()

        oimg = plt.imshow(raster, alpha=0.25, cmap='jet', vmin=vmin, vmax=vmax, extent=extent)

        ax = plt.gca()

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)

        plt.colorbar(oimg, cax=cax)

#        plt.show()

        return fig

    def mask(self, mask, mask_value=True):
        if mask_value is bool:
            self.img = np.ma.masked_where(mask, self.img)
        else:
            self.img = np.ma.masked_where(mask == mask_value, self.img)

    def unmask(self):
        if isinstance(self.img, np.ma.masked_array):
            self.img = self.img.data
        else:
            pass

    def random_points(self, Npts, edge_buffer=None):
        # first, if we don't have an edge buffer and don't have a mask, everything is easy.
        if edge_buffer is None:
            indices = np.arange(self.img.size)  # a list of indices
            if not isinstance(self.img, np.ma.MaskedArray):
                goodinds = indices[np.isfinite(self.img.reshape(-1))]
            else:
                goodinds = indices[np.logical_and(np.invert(self.img.mask).reshape(-1),
                                   np.isfinite(self.img.data.reshape(-1)))]
            return [np.array(np.unravel_index(x, self.img.shape)) for x in random.sample(goodinds, Npts)]
        elif edge_buffer is not None:
            tmp_img = self.img.data[edge_buffer:-edge_buffer, edge_buffer:-edge_buffer]
            indices = np.arange(tmp_img.size)
            if isinstance(self.img, np.ma.MaskedArray):
                tmp_mask = self.img.mask[edge_buffer:-edge_buffer, edge_buffer:-edge_buffer]
                goodinds = indices[np.logical_and(np.invert(tmp_mask).reshape(-1), np.isfinite(tmp_img.reshape(-1)))]
            else:
                goodinds = indices[np.isfinite(tmp_img.reshape(-1))]
            # return a random list as above, but remember to shift everything by the edge buffer.
            return [np.array(np.unravel_index(x, tmp_img.shape))+edge_buffer for x in random.sample(goodinds, Npts)]

    def interp_data(self, xy):
        X, Y = self.xy()
        X = X.reshape(-1)
        Y = Y.reshape(-1)
        tmpImg = self.img.reshape(-1)

        interpData = griddata((X, Y), tmpImg, xy)
        return interpData



import os
import subprocess
import gdal
import numpy as np
import matplotlib.pylab as plt
import scipy.optimize as optimize
from matplotlib.backends.backend_pdf import PdfPages
from mpl_toolkits.axes_grid1 import make_axes_locatable
from pybob.GeoImg import GeoImg
from pybob.image_tools import create_mask_from_shapefile


def false_hillshade(dH, title, pp):
    niceext = np.array([dH.xmin, dH.xmax, dH.ymin, dH.ymax])/1000.
    fig = plt.figure(figsize=(7, 5))
    ax = plt.gca()
    im1 = ax.imshow(dH.img, extent=niceext)
    im1.set_clim(-20, 20)
    im1.set_cmap('Greys')
    fig.suptitle(title, fontsize=14)
    numwid = max([len('%.1f m' % np.nanmean(dH.img)),
                 len('%.1f m' % np.nanmedian(dH.img)), len('%.1f m' % np.nanstd(dH.img))])
    plt.annotate('MEAN:'.ljust(8) + ('%.1f m' % np.nanmean(dH.img)).rjust(numwid), xy=(0.65, 0.89),
                 xycoords='axes fraction', fontsize=12, fontweight='bold', color='red', family='monospace')
    plt.annotate('MEDIAN:'.ljust(8) + ('%.1f m' % np.nanmedian(dH.img)).rjust(numwid),
                 xy=(0.65, 0.80), xycoords='axes fraction', fontsize=12, fontweight='bold',
                 color='red', family='monospace')
    plt.annotate('STD:'.ljust(8) + ('%.1f m' % np.nanstd(dH.img)).rjust(numwid), xy=(0.65, 0.71),
                 xycoords='axes fraction', fontsize=12, fontweight='bold', color='red', family='monospace')

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.05)
    plt.colorbar(im1, cax=cax)

    plt.tight_layout()
    pp.savefig(fig)
    return


def create_stable_mask(img, mask1, mask2):
    # if we have no masks, just return an array of true values
    if mask1 is None and mask2 is None:
        return np.ones(img.img.shape) == 0  # all false, so nothing will get masked.
    elif mask1 is not None and mask2 is None:  # we have a glacier mask, not land
        mask = create_mask_from_shapefile(img, mask1)
        return mask  # returns true where there's glacier, false everywhere else
    elif mask1 is None and mask2 is not None:
        mask = create_mask_from_shapefile(img, mask2)
        return np.logical_not(mask)  # false where there's land, true where there isn't
    else:  # if none of the above, we have two masks.
        gmask = create_mask_from_shapefile(img, mask1)
        lmask = create_mask_from_shapefile(img, mask2)
        return np.logical_or(gmask, np.logical_not(lmask))  # true where there's glacier or water


def preprocess(stable_mask, slope, aspect, master, slave):
    stan = np.tan(np.radians(slope)).astype(np.float32)
    # create a new raster with all the same properties as the masterDEM
    # but replace the raster data with the dH data.
    dH = master.copy(new_raster=(master.img-slave.img).astype(np.float32))
    dH.img[stable_mask] = np.nan

    dHtan = dH.img / stan
    mykeep = ((np.absolute(dH.img) < 60.0) & np.isfinite(dH.img) & (slope > 7.0) & (dH.img != 0.0) & (aspect >= 0))
    dH.img[np.invert(mykeep)] = np.nan
    xdata = aspect[mykeep]
    ydata = dHtan[mykeep]
    sdata = stan[mykeep]

    return dH, xdata, ydata, sdata


def coreg_fitting(xdata, ydata, sdata, title, pp):
    xdata = xdata.astype(np.float64)  # float64 truly necessary?
    ydata = ydata.astype(np.float64)
    sdata = sdata.astype(np.float64)
    # fit using equation 3 of Nuth and Kaeaeb, 2011

    def fitfun(p, x): return p[0] * np.cos(np.radians(p[1] - x)) + p[2]

    def errfun(p, x, y): return fitfun(p, x) - y

    p0 = [-1, 1, -1]
    p1, success = optimize.leastsq(errfun, p0[:], args=(xdata, ydata))
#    print success
#    print p1
    # convert to shift parameters in cartesian coordinates
    xadj = p1[0] * np.sin(np.radians(p1[1]))
    yadj = p1[0] * np.cos(np.radians(p1[1]))
    zadj = p1[2] * sdata.mean(axis=0)

    xp = np.linspace(0, 360, 361)
    yp = fitfun(p1, xp)

    if xdata.size > 50000:
        mysamp = np.random.randint(0, xdata.size, 50000)
    else:
        mysamp = np.arange(0, xdata.size)

    fig = plt.figure(figsize=(7, 5), dpi=600)
    fig.suptitle(title, fontsize=14)
    plt.plot(xdata[mysamp], ydata[mysamp], '.', ms=0.5, color='0.5', rasterized=True)
    plt.plot(xp, np.zeros(xp.size), 'k', ms=3)
    plt.plot(xp, yp, 'r-', ms=2)

    plt.axis([0, 360, -200, 200])
    plt.xlabel('Aspect [degrees]')
    plt.ylabel('dH / tan(slope)')
    numwidth = max([len(' %.1f m' % xadj), len(' %.1f m' % yadj), len(' %.1f m' % zadj)])
    plt.text(20, -125, '$\Delta$x: ' + ('%.1f m' % xadj).rjust(numwidth),
             fontsize=12, fontweight='bold', color='red', family='monospace')
    plt.text(20, -150, '$\Delta$y: ' + ('%.1f m' % yadj).rjust(numwidth),
             fontsize=12, fontweight='bold', color='red', family='monospace')
    plt.text(20, -175, '$\Delta$z: ' + ('%.1f m' % zadj).rjust(numwidth),
             fontsize=12, fontweight='bold', color='red', family='monospace')
    pp.savefig(fig, rasterized=True)

    return xadj, yadj, zadj


def get_geoimg(indata):
    if type(indata) is str or type(indata) is gdal.Dataset:
        return GeoImg(indata)
    elif type(indata) is GeoImg:
        return indata
    else:
        raise TypeError('input data must be a string pointing to a gdal dataset, or a GeoImg object.')


def dem_coregistration(masterDEM, slaveDEM, glaciermask=None, landmask=None, outdir='.'):
    # if the output directory does not exist, create it.
    subprocess.call(["mkdir", "-p", outdir])
    outdir = os.path.abspath(outdir)

    # create the output pdf
    pp = PdfPages(outdir + os.path.sep + 'CoRegistration_Results.pdf')

    if type(masterDEM) is str:
        mfilename = os.path.basename(masterDEM)
    else:
        mfilename = masterDEM.filename

    if type(slaveDEM) is str:
        sfilename = os.path.basename(slaveDEM)
    else:
        sfilename = slaveDEM.filename

    masterDEM = get_geoimg(masterDEM)
    slaveDEM = get_geoimg(slaveDEM)

    masterDEM = masterDEM.reproject(slaveDEM)  # need to resample masterDEM to cell size of slave.

    # get the mask set up
    stable_mask = create_stable_mask(masterDEM, glaciermask, landmask)

    masterDEM.write('tmp_master.tif', out_folder=outdir)
    masterDEM.mask(stable_mask)

    # create slope, aspect rasters
    # slope, aspect = calcSlopeAspect(masterDEM.img, masterDEM.dx)
    callstr = ["gdaldem", "slope", outdir + os.path.sep + 'tmp_master.tif', outdir + os.path.sep + "tmp_slope.tif"]
    slopecall = subprocess.Popen(' '.join(callstr), shell=True,
                                 stdout=open(os.devnull, 'wb'), stderr=subprocess.PIPE).stderr.read()
    print slopecall
    slpin = GeoImg('tmp_slope.tif', in_dir=outdir)
    slpin = slpin.reproject(slaveDEM)  # re-calculate and re-project on each step?
    slope = slpin.img

    callstr = ["gdaldem", "aspect", outdir + os.path.sep + 'tmp_master.tif', outdir + os.path.sep + "tmp_aspect.tif"]
    aspectcall = subprocess.Popen(' '.join(callstr), shell=True,
                                  stdout=open(os.devnull, 'wb'), stderr=subprocess.PIPE).stderr.read()
    print aspectcall
    aspin = GeoImg('tmp_aspect.tif', in_dir=outdir)
    aspin = aspin.reproject(slaveDEM)
    aspect = aspin.img

    # while loop!
    mythresh = np.float64(100)  # float64 really necessary?
    mystd = np.float64(100)
    mycount = 0
    tot_dx = np.float64(0)
    tot_dy = np.float64(0)
    tot_dz = np.float64(0)
    magnthresh = 100
    mytitle = 'DEM difference: pre-coregistration'
    slaves = []
    slaves.append(slaveDEM.reproject(masterDEM))
    slaves[-1].mask(stable_mask)

    while mythresh > 2 and magnthresh > 1:
        if mycount != 0:
            slaves.append(slaves[-1].reproject(masterDEM))
            slaves[-1].mask(stable_mask)
            mytitle = "DEM difference: After Iteration %s" % mycount
        mycount += 1
        print "Running iteration #%s" % mycount

        dH, xdata, ydata, sdata = preprocess(stable_mask, slope, aspect, masterDEM, slaves[-1])
        false_hillshade(dH, mytitle, pp)

        mythresh = 100 * (mystd-np.nanstd(dH.img))/mystd
        mystd = np.nanstd(dH.img)

        mytitle2 = "Co-registration: Iteration %s" % mycount
        dx, dy, dz = coreg_fitting(xdata, ydata, sdata, mytitle2, pp)
        tot_dx += dx
        tot_dy += dy
        tot_dz += dz
        magnthresh = np.sqrt(np.square(dx)+np.square(dy)+np.square(dz))
        print tot_dx, tot_dy, tot_dz
        # print np.nanmean(slaves[-1].img)

        # print slaves[-1].xmin, slaves[-1].ymin

        # shift most recent slave DEM
        slaves[-1].shift(dx, dy)  # shift in x,y
        # print tot_dx, tot_dy
        # no idea why slaves[-1].img += dz doesn't work, but the below seems to.
        slaves[-1] = slaves[-1].copy(new_raster=slaves[-1].img+dz)  # shift in z

        # print np.nanmean(slaves[-1].img)

        slaves[-1] = slaves[-1].reproject(masterDEM)
        # slaves[-1].display()
        if mythresh > 2 and magnthresh > 1:
            dH = None
            dx = None
            dy = None
            dz = None
            xdata = None
            ydata = None
            sdata = None
        else:
            dHfinal = dH
            mytitle = "DEM difference: After Iteration %s" % mycount
            false_hillshade(dH, mytitle, pp)
            # slaves.pop(-1)

    # create new raster with dH sample used for co-registration as the band
    # dHSample = dH.copy(new_raster=dHpost_sample)
    # dHSample.write(outdir + os.path.sep + 'dHpost_sample.tif') # have to fill these in!
    # save full dH output
    dHfinal.write('dHpost.tif', out_folder=outdir)
    # save adjusted slave dem
    if sfilename is not None:
        slaveoutfile = '.'.join(sfilename.split('.')[0:-1]) + '_adj.tif'
    else:
        slaveoutfile = 'slave_adj.tif'
    slaves[-1].write(slaveoutfile, out_folder=outdir)

    if mfilename is not None:
        mastoutfile = '.'.join(mfilename.split('.')[0:-1]) + '_adj.tif'
    else:
        mastoutfile = 'master_adj.tif'
    masterDEM.write(mastoutfile, out_folder=outdir)

    pp.close()
    print "Fin."

    return masterDEM, slaves[-1]
